import pytest

from sleeplab_format.models import *
from sleeplab_format import writer


@pytest.fixture(scope='session')
def subject_ids():
    return ['10001', '10002', '10003']


def subject_metadata(sid):
    return {
        'subject_id': sid,
        'recording_start_ts': '2018-01-01T23:10:04',
        'age': 24.0,
        'bmi': 25.56,
        'sex': 'MALE'
    }


def sample_arrays():
    return {
        's1': {
            'attributes': {
                'name': 's1',
                'start_ts': '2018-01-01T23:10:04',
                'sampling_rate': 32.0,
                'unit': 'V'
            },
            'values': 0.123 * np.ones(60*32, dtype=np.float32),
        },
        's2': {
            'attributes': {
                'name': 's2',
                'start_ts': '2018-01-01T23:10:04',
                'sampling_rate': 64.0,
                'unit': 'mV'
            },
            'values': 1.23 * np.ones(60*64, dtype=np.float32),
        },
    }


def events():
    return {
        'scorer': 'automatic',
        'type': 'aasmevents',
        'annotations': [
        {
            'name': 'SPO2_DESAT',
            'start_ts': '2018-01-01T23:10:04',
            'start_sec': 0.0,
            'duration': 15.0,
            'extra_attributes': {
                'min_spo2': 93,
                'desat_limit': 4
            }
        },
        {
            'name': 'APNEA_CENTRAL',
            'start_ts': '2018-01-01T23:10:24',
            'start_sec': 20.0,
            'duration': 10.0
        },
        {
            'name': 'HYPOPNEA',
            'start_ts': '2018-01-01T23:10:34',
            'start_sec': 30.0,
            'duration': 10.0,
        }],
    }


def hypnogram():
    return {
        'scorer': 'scorer_1',
        'type': 'hypnogram',
        'annotations': [
        {
            'name': 'N1',
            'start_ts': '2018-01-01T23:10:04',
            'start_sec': 0.0,
            'duration': 30.0,
        },
        {
            'name': 'W',
            'start_ts': '2018-01-01T23:10:34',
            'start_sec': 30.0,
            'duration': 30.0
        },
        ],
    }


def study_logs():
    return {
        'scorer': 'study',
        'type': 'logs',
        'annotations': [
        {
            'name': 'F3 - Impedance 1,4k',
            'start_ts': '2018-01-01T23:10:04',
            'start_sec': 0.0,
            'duration': 0.0   
        },
        {
            'name': 'LIGHTS OUT',
            'start_ts': '2018-01-01T23:10:04',
            'start_sec': 0.0,
            'duration': 0.0   
        },
        {
            'name': 'LIGHTS ON',
            'start_ts': '2018-01-01T23:11:04',
            'start_sec': 60.0,
            'duration': 0.0   
        }
    ]}


@pytest.fixture
def subjects(subject_ids):
    subjs = {}
    for sid in subject_ids:
        metadata = SubjectMetadata(**subject_metadata(sid))
        
        dict_arrays = sample_arrays()
        arrays = {
            k: SampleArray(
                attributes=ArrayAttributes(**v['attributes']),
                values_func=lambda v=v: v['values'])
            for k, v in dict_arrays.items()
        }
        subjs[sid] = Subject(
            metadata=metadata,
            sample_arrays=arrays,
            annotations={
                'automatic_aasmevents': AASMEvents.model_validate(events()),
                'scorer_1_hypnogram': Hypnogram.model_validate(hypnogram()),
                'study_logs': Logs.model_validate(study_logs())
            }
        )

    return subjs


@pytest.fixture
def series(subjects):
    series = Series(
        name='series1',
        subjects=subjects
    )
    return {'series1': series}


@pytest.fixture
def dataset(series):
    dataset = Dataset(
        name='dataset1',
        series=series
    )
    return dataset

@pytest.fixture
def ds_dir(dataset, tmp_path):
    basedir = tmp_path / 'datasets'
    writer.write_dataset(dataset, basedir)

    return basedir / dataset.name



import pytest
import shutil

from sleeplab_format.models import *
from pathlib import Path

from .create_datasets import create_datasets
from sleeplab_format.test_utils.fixtures import *


@pytest.fixture(scope='session', autouse=True)
def recreate_test_files() -> None:
    """Recreate the separate test data files at the beginning of tests."""
    # First remove the old files if existing
    ds_dir = Path(__file__).parent / 'datasets'
    shutil.rmtree(ds_dir, ignore_errors=True)

    # Then recreate them
    create_datasets(ds_dir)



import subprocess

from sleeplab_format import reader, writer
from pathlib import Path

from sleeplab_format.models import Dataset


def _assert_dirs_equal(dir1, dir2):
    p = subprocess.run(['diff', '-r', dir1, dir2])
    assert p.returncode == 0


def _assert_datasets_equal(ds1, ds2):
    """Compare the equality of two Dataset objects.
    
    Cannot just compare ds1 == ds2 since the 
    SampleArray.values_func objects can be different.
    """
    assert ds1.name == ds2.name
    assert ds1.series.keys() == ds2.series.keys()
    for (k, series1) in ds1.series.items():
        series2 = ds2.series[k]
        assert series1.name == series2.name
        assert series1.subjects.keys() == series2.subjects.keys()

        for (sk, subj1) in series1.subjects.items():
            subj2 = series2.subjects[sk]
            assert subj1.metadata == subj2.metadata
            # In Pydantic v2, a subclass of a model cannot be equal to the superclass,
            # so need to dump the model to dicts
            annotations1 = {k: v.model_dump() for k, v in subj1.annotations.items()}
            annotations2 = {k: v.model_dump() for k, v in subj2.annotations.items()}
            assert annotations1 == annotations2
            assert subj1.sample_arrays.keys() == subj2.sample_arrays.keys()
            
            for (arr_k, arr1) in subj1.sample_arrays.items():
                arr2 = subj2.sample_arrays[arr_k]
                assert arr1.attributes == arr2.attributes
                assert (arr1.values == arr2.values).all()


def test_read_dataset(dataset: Dataset):
    tests_ds_dir = Path(__file__).parent / 'datasets' / 'dataset1'
    ds_read = reader.read_dataset(tests_ds_dir)

    _assert_datasets_equal(dataset, ds_read)


def test_read_ds_specify_series(dataset: Dataset):
    tests_ds_dir = Path(__file__).parent / 'datasets' / 'dataset1'
    ds_read = reader.read_dataset(tests_ds_dir, series_names=['series1'])

    _assert_datasets_equal(dataset, ds_read)


def test_write_read(dataset: Dataset, tmp_path: Path):
    ds_dir = tmp_path / 'datasets'
    writer.write_dataset(dataset, ds_dir)
    ds_read = reader.read_dataset(ds_dir / dataset.name)
    _assert_datasets_equal(dataset, ds_read)


def test_write_read_parquet(dataset: Dataset, tmp_path: Path):
    ds_dir = tmp_path / 'datasets'
    writer.write_dataset(dataset, ds_dir,
                         annotation_format='parquet',
                         array_format='parquet')

    ds_read = reader.read_dataset(ds_dir / dataset.name)
    _assert_datasets_equal(dataset, ds_read)


def test_read_write(tmp_path: Path):
    tests_ds_dir = Path(__file__).parent / 'datasets'
    ds_read = reader.read_dataset(tests_ds_dir / 'dataset1')

    ds_dir = tmp_path / 'datasets'
    writer.write_dataset(ds_read, ds_dir)

    # Assert that the created dataset is equal to tests/datasets
    _assert_dirs_equal(str(ds_dir.resolve()), str(tests_ds_dir.resolve()))



import argparse
import logging
import numpy as np
import pandas as pd
import sleeplab_format as slf

from datetime import datetime, timedelta
from functools import partial
from pathlib import Path
from pyedflib import highlevel
from typing import Any


logger = logging.getLogger(__name__)


def parse_hypnogram(ann_header: dict[str, Any], scorer: str) -> slf.models.RKHypnogram:
    """Parse the sleep stages according to Rechtschaffen and Kales."""
    stage_map = {
        'Sleep stage W': slf.models.RKSleepStage.W,
        'Sleep stage 1': slf.models.RKSleepStage.S1,
        'Sleep stage 2': slf.models.RKSleepStage.S2,
        'Sleep stage 3': slf.models.RKSleepStage.S3,
        'Sleep stage 4': slf.models.RKSleepStage.S4,
        'Sleep stage R': slf.models.RKSleepStage.R,
        'Movement time': slf.models.RKSleepStage.MOVEMENT,
        'Sleep stage ?': slf.models.RKSleepStage.UNSCORED,
    }
    annotations = [
        slf.models.Annotation[slf.models.RKSleepStage](
            name=stage_map[stage_str],
            start_ts=ann_header['startdate'] + timedelta(seconds=start_sec),
            start_sec=start_sec,
            duration=float(duration)
        )
        for start_sec, duration, stage_str in ann_header['annotations']
    ]

    return slf.models.RKHypnogram(scorer=scorer, annotations=annotations)


def read_signal_from_edf(edf_path: Path, ch_name, dtype=np.float32) -> np.array:
    """Read a single signal from an EDF file."""
    s, _, _ = highlevel.read_edf(str(edf_path), ch_names=ch_name)
    assert s.ndim == 2 and s.shape[0] == 1
    return s[0].astype(dtype)


def parse_sample_arrays(psg_path: Path) -> dict[str, slf.models.SampleArray]:
    """Parse all sample arrays from a PSG EDF file."""
    sample_arrays = {}
    headers = highlevel.read_edf_header(str(psg_path))
    for sig_header in headers['SignalHeaders']:
        # We define a function that loads and returns the signal
        # to avoid reading all signals of all subjects into memory at the same time
        s_load_func = partial(read_signal_from_edf, edf_path=psg_path, ch_name=sig_header['label'])
        array_attributes = slf.models.ArrayAttributes(
            name=sig_header['label'],
            start_ts=headers['startdate'],
            sampling_rate=sig_header['sample_frequency'],
            unit=sig_header['dimension'],
            sensor_info=sig_header['transducer'],
            amplifier_info=sig_header['prefilter']
        )
        sarr = slf.models.SampleArray(attributes=array_attributes, values_func=s_load_func)
        sample_arrays[sig_header['label']] = sarr

    return sample_arrays


def resolve_lightsoff(d, t):
    """Figure out if the date of lights off is the same as start_ts or the next."""
    if t >= d.time():
        return datetime.combine(d.date(), t)
    else:
        return datetime.combine((d + timedelta(days=1)).date(), t)


def parse_subject(src_dir: Path, meta: pd.Series) -> slf.models.Subject:
    """Parse the PSG recording and annotations for a single Subject."""
    # Parse recording
    sc_dir = src_dir / 'sleep-cassette'
    psg_path = next(sc_dir.glob(f'SC4{meta.subject:02d}{meta.night}*PSG.edf'))
    sample_arrays = parse_sample_arrays(psg_path)

    # Parse annotations
    # The eight letter of annotation file name is the scorer id
    ann_path = next(sc_dir.glob(f'SC4{meta.subject:02d}{meta.night}*Hypnogram.edf'))
    scorer = ann_path.name[7]
    ann_header = highlevel.read_edf_header(str(ann_path))
    hg = parse_hypnogram(ann_header, scorer)

    # Need to parse lights off using start_ts, since meta.LightsOff has only time, not date
    start_ts = ann_header['startdate']
    lights_off = resolve_lightsoff(start_ts, meta.LightsOff)

    metadata = slf.models.SubjectMetadata(
        subject_id=str(meta.subject),
        recording_start_ts=start_ts,
        lights_off=lights_off,
        age=meta.age,
        sex=slf.models.Sex.FEMALE if meta['sex (F=1)'] == 1 else slf.models.Sex.MALE
    )

    return slf.models.Subject(
        metadata=metadata,
        sample_arrays=sample_arrays,
        annotations={f'{hg.scorer}_{hg.type}': hg}
    )


def convert_data(
        src_dir: Path,
        dst_dir: Path,
        annotation_format: str,
        array_format: str,
        clevel: int) -> None:
    """Convert the sleep-cassette dataset to sleeplab format."""    
    # SC-subjects.xls contains ID, night, age, sex, and lights off time for each subject.
    excel_path = src_dir / 'SC-subjects.xls'
    meta_df = pd.read_excel(excel_path)

    # Sleep-cassette consists of two 20h PSGs recorded on consecutive days (1 and 2).
    all_subjects = {1: {}, 2: {}}
    for _, s in meta_df.iterrows():
        subject = parse_subject(src_dir, s)
        all_subjects[s.night][str(s.subject)] = subject
        
    night1_series = slf.models.Series(name='sc-night1', subjects=all_subjects[1])
    night2_series = slf.models.Series(name='sc-night2', subjects=all_subjects[2])
    ds = slf.models.Dataset(
        name='sleep-cassette',
        series={'sc-night1': night1_series, 'sc-night2': night2_series})

    slf.writer.write_dataset(ds, dst_dir, annotation_format=annotation_format,
                             array_format=array_format, compression_level=clevel)


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-s', '--src-dir', help='Path to the sleep-edfx root folder, e.g. `sleep-edfx/1.0.0`')
    parser.add_argument(
        '-d', '--dst-dir', help='Path to the folder where the sleeplab-format dataset is saved.')
    parser.add_argument('--array-format', default='numpy', help='Array format, `numpy` or `parquet`')
    parser.add_argument('--annotation-format', default='json', help='Annotation format, `json` or `parquet`')
    parser.add_argument('--clevel', type=int, default=9, help='Compression level for the numerical arrays.')
    
    return parser


if __name__ == '__main__':
    parser = _create_parser()
    args = parser.parse_args()

    convert_data(
        src_dir=Path(args.src_dir).expanduser(),
        dst_dir=Path(args.dst_dir).expanduser(),
        annotation_format=args.annotation_format,
        array_format=args.array_format,
        clevel=args.clevel
    )



"""Download Dreem datasets. Modified from
https://github.com/Dreem-Organization/dreem-learning-open/blob/master/download_data.py
"""
import argparse
import boto3
import tqdm

from botocore import UNSIGNED
from botocore.client import Config
from pathlib import Path


def download_data(dst_dir: Path):
    client = boto3.client('s3', config=Config(signature_version=UNSIGNED))

    bucket_objects = client.list_objects(Bucket='dreem-dod-o')["Contents"]
    print("\n Downloading H5 files and annotations from S3 for DOD-O")
    savedir = dst_dir / 'dodo'
    savedir.mkdir(exist_ok=True)
    for bucket_object in tqdm.tqdm(bucket_objects):
        filename = bucket_object["Key"]
        savepath = savedir / filename
        client.download_file(
            Bucket="dreem-dod-o",
            Key=filename,
            Filename=savepath 
        )
        
    bucket_objects = client.list_objects(Bucket='dreem-dod-h')["Contents"]
    print("\n Downloading H5 files and annotations from S3 for DOD-H")
    savedir = dst_dir / 'dodh'
    savedir.mkdir(exist_ok=True)
    for bucket_object in tqdm.tqdm(bucket_objects):
        filename = bucket_object["Key"]
        savepath = savedir / filename
        client.download_file(
            Bucket="dreem-dod-h",
            Key=filename,
            Filename=savepath
        )


def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--dst-dir',
                        help='the directory where to download the data')
    return parser


if __name__ == '__main__':
    parser = create_parser()
    args = parser.parse_args()
    dst_dir = Path(args.dst_dir)
    download_data(dst_dir=dst_dir)



import argparse
import h5py
import json
import logging
import numpy as np
import pyedflib
import sleeplab_format as slf
import subprocess
import time

from pathlib import Path


logger = logging.getLogger(__name__)


def profile_time(f, *args, **kwargs):
    t1 = time.time()
    p1 = time.process_time()
    res = f(*args, **kwargs)
    t2 = time.time()
    p2 = time.process_time()
    
    return {
        'time': t2 - t1,
        'process_time': p2 - p1,
        'result': res
    }


def get_size(dir: Path):
    """Get folder size in bytes."""
    res = subprocess.run(['du', '-sb', str(dir)], capture_output=True, text=True)
    return res.stdout.split()[0]


def convert_dod(src_dir: Path, dst_dir: Path, array_format: str, clevel: str) -> None:
    subprocess.run(['python', '../dod_sleep_staging/convert_data.py',
                    '-s', src_dir,
                    '-d', dst_dir,
                    '--array-format', array_format,
                    '--clevel', clevel])


def iterate_dod(src_dir: Path) -> None:
    """Iterate over dodo and dodh h5 files and calculate mean.

    h5py.File with driver='core' loads the whole file into memory.
    """
    logger.info(f'Iterating {str(src_dir)}')
    for series in ['dodo', 'dodh']:
        for h5_path in (src_dir / series).iterdir():
            logger.info(f'Reading {str(h5_path)}')
            h5 = h5py.File(h5_path, 'r')
            for signal_type in h5['signals']:
                for signal_name in h5['signals'][signal_type]:
                    s = h5['signals'][signal_type][signal_name][:].astype(np.float32)
                    _ = np.mean(s)
            h5.close()


def convert_sc(src_dir: Path, dst_dir: Path, array_format: str, clevel: str) -> None:
    subprocess.run(['python', '../sleep_cassette_conversion/convert_data.py',
                    '-s', src_dir,
                    '-d', dst_dir,
                    '--array-format', array_format,
                    '--clevel', clevel])


def iterate_sc(src_dir: Path) -> None:
    logger.info(f'Iterating {str(src_dir)}')
    for edf_path in (src_dir / 'sleep-cassette').glob('*.edf'):
        sigs, sig_headers, header = pyedflib.highlevel.read_edf(str(edf_path))
        for s in sigs:
            _ = np.mean(s)


def iterate_slf_ds(ds_dir: Path) -> None:
    logger.info(f'Iterating {str(ds_dir)}')
    ds = slf.reader.read_dataset(ds_dir)
    for series in ds.series.values():
        for subject in series.subjects.values():
            for sarr in subject.sample_arrays.values():
                # Read SampleArray values to memory
                s = sarr.values_func()[:]
                _ = np.mean(s)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--slf-dir', help='Root folder where the converted datasets will be stored.')
    parser.add_argument('--output-path', help='Path to the benchmark result JSON file.')
    parser.add_argument('--dod-orig-dir', help='Path to the downloaded Dreem data')
    parser.add_argument('--sc-orig-dir', help='Path to the downloaded Sleep-Cassette data')
    args = parser.parse_args()
    slf_dir = Path(args.slf_dir)
    output_path = Path(args.output_path)

    res = {'dod': {}, 'sc': {}}

    ######################
    # Dreem open datasets
    ######################

    dod_orig_dir = Path(args.dod_orig_dir)
    dod_slf_numpy_dir = slf_dir / 'dod_slf_numpy'
    dod_slf_zarr_clevel_3_dir = slf_dir / 'dod_slf_clevel_3'
    dod_slf_zarr_clevel_9_dir = slf_dir / 'dod_slf_clevel_9'
    dod_slf_zarr_clevel_22_dir = slf_dir / 'dod_slf_clevel_22'

    # Original
    res['dod']['original'] = {}
    res['dod']['original']['size'] = int(get_size(dod_orig_dir)) / 1e9

    iterate_res = profile_time(iterate_dod, dod_orig_dir)
    res['dod']['original']['read_time'] = iterate_res['time']
    res['dod']['original']['read_process_time'] = iterate_res['process_time']

    # NumPy
    convert_res = profile_time(convert_dod, dod_orig_dir, dod_slf_numpy_dir, 'numpy', '-1')
    res['dod']['numpy'] = {}
    res['dod']['numpy']['size'] = int(get_size(dod_slf_numpy_dir)) / 1e9
    res['dod']['numpy']['write_time'] = convert_res['time']
    res['dod']['numpy']['write_process_time'] = convert_res['process_time']

    iterate_res = profile_time(iterate_slf_ds, dod_slf_numpy_dir / dod_orig_dir.name)
    res['dod']['numpy']['read_time'] = iterate_res['time']
    res['dod']['numpy']['read_process_time'] = iterate_res['process_time']

    # Zarr compression level 3
    convert_res = profile_time(convert_dod, dod_orig_dir, dod_slf_zarr_clevel_3_dir, 'zarr', '3')
    res['dod']['zarr_clevel_3'] = {}
    res['dod']['zarr_clevel_3']['size'] = int(get_size(dod_slf_zarr_clevel_3_dir)) / 1e9
    res['dod']['zarr_clevel_3']['write_time'] = convert_res['time']
    res['dod']['zarr_clevel_3']['write_process_time'] = convert_res['process_time']

    iterate_res = profile_time(iterate_slf_ds, dod_slf_zarr_clevel_3_dir / dod_orig_dir.name)
    res['dod']['zarr_clevel_3']['read_time'] = iterate_res['time']
    res['dod']['zarr_clevel_3']['read_process_time'] = iterate_res['process_time']
    
    # Zarr compression level 9
    convert_res = profile_time(convert_dod, dod_orig_dir, dod_slf_zarr_clevel_9_dir, 'zarr', '9')
    res['dod']['zarr_clevel_9'] = {}
    res['dod']['zarr_clevel_9']['size'] = int(get_size(dod_slf_zarr_clevel_9_dir)) / 1e9
    res['dod']['zarr_clevel_9']['write_time'] = convert_res['time']
    res['dod']['zarr_clevel_9']['write_process_time'] = convert_res['process_time']

    iterate_res = profile_time(iterate_slf_ds, dod_slf_zarr_clevel_9_dir / dod_orig_dir.name)
    res['dod']['zarr_clevel_9']['read_time'] = iterate_res['time']
    res['dod']['zarr_clevel_9']['read_process_time'] = iterate_res['process_time']

    # Zarr compression level 22
    convert_res = profile_time(convert_dod, dod_orig_dir, dod_slf_zarr_clevel_22_dir, 'zarr', '22')
    res['dod']['zarr_clevel_22'] = {}
    res['dod']['zarr_clevel_22']['size'] = int(get_size(dod_slf_zarr_clevel_22_dir)) / 1e9
    res['dod']['zarr_clevel_22']['write_time'] = convert_res['time']
    res['dod']['zarr_clevel_22']['write_process_time'] = convert_res['process_time']

    iterate_res = profile_time(iterate_slf_ds, dod_slf_zarr_clevel_22_dir / dod_orig_dir.name)
    res['dod']['zarr_clevel_22']['read_time'] = iterate_res['time']
    res['dod']['zarr_clevel_22']['read_process_time'] = iterate_res['process_time']

    #################
    # Sleep-Cassette
    #################

    sc_orig_dir = Path(args.sc_orig_dir)
    sc_slf_numpy_dir = slf_dir / 'sc_slf_numpy'
    sc_slf_zarr_clevel_3_dir = slf_dir / 'sc_slf_clevel_3'
    sc_slf_zarr_clevel_9_dir = slf_dir / 'sc_slf_clevel_9'
    sc_slf_zarr_clevel_22_dir = slf_dir / 'sc_slf_clevel_22'

    # Original
    res['sc']['original'] = {}
    res['sc']['original']['size'] = int(get_size(sc_orig_dir / 'sleep-cassette')) / 1e9

    iterate_res = profile_time(iterate_sc, sc_orig_dir)
    res['sc']['original']['read_time'] = iterate_res['time']
    res['sc']['original']['read_process_time'] = iterate_res['process_time']

    # NumPy
    convert_res = profile_time(convert_sc, sc_orig_dir, sc_slf_numpy_dir, 'numpy', '-1')
    res['sc']['numpy'] = {}
    res['sc']['numpy']['size'] = int(get_size(sc_slf_numpy_dir)) / 1e9
    res['sc']['numpy']['write_time'] = convert_res['time']
    res['sc']['numpy']['write_process_time'] = convert_res['process_time']

    iterate_res = profile_time(iterate_slf_ds, sc_slf_numpy_dir / 'sleep-cassette')
    res['sc']['numpy']['read_time'] = iterate_res['time']
    res['sc']['numpy']['read_process_time'] = iterate_res['process_time']

    # Zarr compression level 3
    convert_res = profile_time(convert_sc, sc_orig_dir, sc_slf_zarr_clevel_3_dir, 'zarr', '3')
    res['sc']['zarr_clevel_3'] = {}
    res['sc']['zarr_clevel_3']['size'] = int(get_size(sc_slf_zarr_clevel_3_dir)) / 1e9
    res['sc']['zarr_clevel_3']['write_time'] = convert_res['time']
    res['sc']['zarr_clevel_3']['write_process_time'] = convert_res['process_time']

    iterate_res = profile_time(iterate_slf_ds, sc_slf_zarr_clevel_3_dir / 'sleep-cassette')
    res['sc']['zarr_clevel_3']['read_time'] = iterate_res['time']
    res['sc']['zarr_clevel_3']['read_process_time'] = iterate_res['process_time']
    
    # Zarr compression level 9
    convert_res = profile_time(convert_sc, sc_orig_dir, sc_slf_zarr_clevel_9_dir, 'zarr', '9')
    res['sc']['zarr_clevel_9'] = {}
    res['sc']['zarr_clevel_9']['size'] = int(get_size(sc_slf_zarr_clevel_9_dir)) / 1e9
    res['sc']['zarr_clevel_9']['write_time'] = convert_res['time']
    res['sc']['zarr_clevel_9']['write_process_time'] = convert_res['process_time']

    iterate_res = profile_time(iterate_slf_ds, sc_slf_zarr_clevel_9_dir / 'sleep-cassette')
    res['sc']['zarr_clevel_9']['read_time'] = iterate_res['time']
    res['sc']['zarr_clevel_9']['read_process_time'] = iterate_res['process_time']

    # Zarr compression level 22
    convert_res = profile_time(convert_sc, sc_orig_dir, sc_slf_zarr_clevel_22_dir, 'zarr', '22')
    res['sc']['zarr_clevel_22'] = {}
    res['sc']['zarr_clevel_22']['size'] = int(get_size(sc_slf_zarr_clevel_22_dir)) / 1e9
    res['sc']['zarr_clevel_22']['write_time'] = convert_res['time']
    res['sc']['zarr_clevel_22']['write_process_time'] = convert_res['process_time']

    iterate_res = profile_time(iterate_slf_ds, sc_slf_zarr_clevel_22_dir / 'sleep-cassette')
    res['sc']['zarr_clevel_22']['read_time'] = iterate_res['time']
    res['sc']['zarr_clevel_22']['read_process_time'] = iterate_res['process_time']

    # Save results
    with open(args.output_path, 'w') as f:
        json.dump(res, f, indent=2)
    



"""Data type definitions for the sleeplab format."""
import numpy as np
import zarr

from collections.abc import Callable
from datetime import datetime
from enum import Enum
from functools import cached_property
from pydantic import BaseModel, Field, model_validator
from pydantic.functional_validators import AfterValidator
from typing import Any, Generic, Literal, Optional, TypeVar
from typing_extensions import Annotated
from .version import __version__


SLEEPLAB_FORMAT_VERSION = __version__


def unset_tzinfo(v: datetime):
    return v.replace(tzinfo=None)


NaiveDatetime = Annotated[datetime, AfterValidator(unset_tzinfo)]


class Sex(str, Enum):
    FEMALE = 'FEMALE'
    MALE = 'MALE'
    OTHER = 'OTHER'


class AASMSleepStage(str, Enum):
    """Sleep stages according to the AASM manual v2.6."""
    W = 'W'
    N1 = 'N1'
    N2 = 'N2'
    N3 = 'N3'
    R = 'R'

    # These are not part of the AASM v2.6 definition, but still commonly used
    UNSURE = 'UNSURE'
    UNSCORED = 'UNSCORED'
    ARTIFACT = 'ARTIFACT'


class RKSleepStage(str, Enum):
    """Sleep stages according to the Rechtschaffen and Kales manual.

    Stages 1, 2, 3, and 4 are prefixed with `S`.
    """
    W = 'W'
    S1 = 'S1'
    S2 = 'S2'
    S3 = 'S3'
    S4 = 'S4'
    R = 'R'

    # TODO: Are these part of the original manual?
    MOVEMENT = 'MOVEMENT'
    UNSCORED = 'UNSCORED'


class AASMEvent(str, Enum):
    """Enum for events scored according to the AASM manual v2.6."""
    # A generic class for artifacts
    ARTIFACT = 'ARTIFACT'

    # A generic class for unsure
    UNSURE = 'UNSURE'

    # Arousal events
    AROUSAL = 'AROUSAL'
    AROUSAL_RES = 'AROUSAL_RES'
    AROUSAL_SPONT = 'AROUSAL_SPONT'
    AROUSAL_LM = 'AROUSAL_LM'
    AROUSAL_PLM = 'AROUSAL_PLM'
    RERA = 'RERA'

    # Cardiac events
    # TODO: Add these if a use case ever pops up

    # Movement events
    BRUXISM = 'BRUXISM'
    LM = 'LM'  # Leg movement
    LM_LEFT = 'LM_LEFT',
    LM_RIGHT = 'LM_RIGHT',
    PLM = 'PLM'  # Periodic leg movement
    PLM_LEFT = 'PLM_LEFT'
    PLM_RIGHT = 'PLM_RIGHT'
    
    # Respiratory events
    APNEA = 'APNEA'
    APNEA_CENTRAL = 'APNEA_CENTRAL'
    APNEA_OBSTRUCTIVE = 'APNEA_OBSTRUCTIVE'
    APNEA_MIXED = 'APNEA_MIXED'
    HYPOPNEA = 'HYPOPNEA'
    HYPOPNEA_CENTRAL = 'HYPOPNEA_CENTRAL'
    HYPOPNEA_OBSTRUCTIVE = 'HYPOPNEA_OBSTRUCTIVE'
    SPO2_DESAT = 'SPO2_DESAT'
    SNORE = 'SNORE'


class SubjectMetadata(BaseModel, extra='forbid'):
    subject_id: str

    # Recording start time
    recording_start_ts: NaiveDatetime

    lights_off: Optional[NaiveDatetime] = None
    lights_on: Optional[NaiveDatetime] = None
    analysis_start: Optional[NaiveDatetime] = None
    analysis_end: Optional[NaiveDatetime] = None

    age: Optional[float] = None
    bmi: Optional[float] = None
    sex: Optional[Sex] = None

    additional_info: Optional[dict[str, Any]] = None
    

class ArrayAttributes(BaseModel, extra='forbid'):
    name: str
    start_ts: NaiveDatetime
    sampling_rate: Optional[float] = None
    sampling_interval: Optional[float] = None
    unit: Optional[str] = None

    amplifier_info: Optional[str] = None
    sensor_info: Optional[str] = None

    # An optional mapping from the array values to some other values,
    # e.g.  {0: 'off', 1: 'on'}
    value_map: Optional[dict[int, str | int]] = None

    @model_validator(mode='after')
    def require_rate_or_interval(self):
        if self.sampling_interval is None:
            _msg = 'either sampling_rate or sampling_interval needs to be defined'
            assert self.sampling_rate is not None, _msg
        else:
            _msg = 'cannot define both sampling_rate and sampling_interval'
            assert self.sampling_rate is None, _msg

        return self


class SampleArray(
        BaseModel,
        extra='forbid',
        ignored_types=(cached_property,)):
    """A pydantic model representing a numerical array with attributes.

    When writing data to sleeplab format, use `values_func` to access the array
    to avoid caching.

    When reading data in sleeplab format, use `values` since the `values_func`
    returned by the reader should return `np.memmap` instead of the full array.
    """
    attributes: ArrayAttributes
    values_func: Callable[[], np.ndarray]
    
    @cached_property
    def values(self) -> np.ndarray | zarr.Array:
        """Use @cached_property so that values_func gets evaluated once
        when values is accessed first time.
        """
        return self.values_func()


AnnotationT = TypeVar('AnnotationT', bound=str)


class Annotation(BaseModel, Generic[AnnotationT], extra='forbid'):
    # Freely defined name of the event
    name: AnnotationT

    # Start time as timestamp
    start_ts: NaiveDatetime

    # Start time as seconds from the start of recording
    start_sec: float

    # Duration of the event in seconds
    # 0.0 if the event is a point in time
    duration: float = 0.0

    # Name of the channel used to annotate the event
    input_channel: Optional[str] = None

    # Freely defined extra attributes, such as desaturation depth
    extra_attributes: Optional[dict[str, Any]] = None


class BaseAnnotations(BaseModel):
    scorer: str
    type: str
    annotations: list[Annotation]


class Annotations(BaseAnnotations):
    type: Literal['annotations'] = 'annotations'
    annotations: list[Annotation[str]]


class Hypnogram(BaseAnnotations):
    """A hypnogram is Annotations consisting of sleep stages."""
    type: Literal['hypnogram'] = 'hypnogram'
    annotations: list[Annotation[AASMSleepStage]]


class RKHypnogram(BaseAnnotations):
    """Hypnogram scored with R&K rules."""
    type: Literal['rkhypnogram'] = 'rkhypnogram'
    annotations: list[Annotation[RKSleepStage]]


class AASMEvents(BaseAnnotations):
    type: Literal['aasmevents'] = 'aasmevents'
    annotations: list[Annotation[AASMEvent]]


class Logs(BaseAnnotations):
    type: Literal['logs'] = 'logs'
    annotations: list[Annotation[str]]


class Subject(BaseModel, extra='forbid'):
    metadata: SubjectMetadata
    sample_arrays: Optional[dict[str, SampleArray]] = Field(None, repr=False)
    annotations: Optional[dict[str, BaseAnnotations]] = Field(None, repr=False)


class Series(BaseModel, extra='forbid'):
    name: str
    subjects: dict[str, Subject] = Field(repr=False)


class Dataset(BaseModel, extra='forbid'):
    name: str
    version: str = SLEEPLAB_FORMAT_VERSION
    series: Optional[dict[str, Series]] = None



"""Methods for reading and splitting sleeplab format data to tf.data.Dataset."""
import logging
import numpy as np
import sleeplab_format as slf
import sleeplab_tf_dataset as sds
import tensorflow as tf

from pathlib import Path
from typing import Any


logger = logging.getLogger(__name__)


component_config = {
    'c3m2': {
        'src_name': 'C3_M2_64Hz',
        'ctype': 'sample_array',
        'fs': 64.0
    },
    'eog2': {
        'src_name': 'EOG2_64Hz',
        'ctype': 'sample_array',
        'fs': 64.0
    },
    'emg': {
        'src_name': 'EMG_64Hz',
        'ctype': 'sample_array',
        'fs': 64.0
    },
    'hypnogram': {
        'src_name': 'manual_consensus_hypnogram.a.parquet',
        'ctype': 'annotation',
        'sampling_interval': 30,
        'return_type': 'segmentation_combined',
        'value_map': {
            'W': 0,
            'N1': 1,
            'N2': 2,
            'N3': 3,
            'R': 4,
            '_default': 0  # Stages unspecified in value_map will be assigned as wake
        }
    }
}


dataset_config = {
    'dodh': {
        'ds_dir': '/tmp/slf/dod_extracted',
        'series_name': 'dodh',
        'roi_src_type': 'annotation',
        'roi_src_name': 'manual_consensus_hypnogram',
        'components': component_config,
        'splits': {
            # Total number of recordings = 25
            'train': {
                'n': 10,
                'config': {
                    'start_sec': -1.0,  # Randomly sample start_sec at each iteration
                    'duration': 60*60,  # Frame length in seconds
                }
            },
            'val': {
                'n': 5,
                'config': {
                    'start_sec': -1.0,
                    'duration': 60*60
                }
            },
            'test': {
                'n': 10,
                'config': {
                    # Use whole-night signals
                    'start_sec': 0.0,
                    'duration': -1.0
                }
            }
        }
    },
    'dodo': {
        'ds_dir': '/tmp/slf/dod_extracted',
        'series_name': 'dodo',
        'roi_src_type': 'annotation',
        'roi_src_name': 'manual_consensus_hypnogram',
        'components': component_config,
        'splits': {
            # Total number of recordings = 56
            'train': {
                'n': 40,
                'config': {
                    'start_sec': -1.0,  # Randomly sample start_sec at each iteration
                    'duration': 60*60,  # Frame length in seconds
                }
            },
            'val': {
                'n': 6,
                'config': {
                    'start_sec': -1.0,
                    'duration': 60*60
                }
            },
            'test': {
                'n': 10,
                'config': {
                    # Use whole-night signals
                    'start_sec': 0.0,
                    'duration': -1.0
                }
            }
        }
    }
}


def parse_dataset_configs(cfg: dict[str, Any]) -> dict[str, sds.config.DatasetConfig]:
    """Parse dict configs to sleeplab_tf_dataset.config.DatasetConfig items."""
    splits = cfg.pop('splits')
    res = {}
    for k, split in splits.items():
        _cfg_dict = cfg.copy()
        _cfg_dict.update(split['config'])
        _cfg = sds.config.DatasetConfig.parse_obj(_cfg_dict)
        res[k] = _cfg

    return res


def load_dataset(cfg: dict[str, Any], seed: int) -> dict[str, tf.data.Dataset]:
    """Load a single series to tf.data.Dataset."""
    logger.info(f"Reading the series {cfg['series_name']}...")
    slf_ds = slf.reader.read_dataset(
        ds_dir=Path(cfg['ds_dir']),
        series_names=[cfg['series_name']]
    )

    logger.info('Creating the splits...')
    # Read the subject IDs
    subj_ids = list(slf_ds.series[cfg['series_name']].subjects.keys())

    # Assert that the sum of split sizes matches the total number of subjects
    sum_split_sizes = sum([split['n'] for split in cfg['splits'].values()])
    _msg = f'Sum of split sizes does not match dataset size ({sum_split_sizes} != {len(subj_ids)})'
    assert sum_split_sizes == len(subj_ids), _msg

    # Randomly shuffle before splitting
    rng = np.random.default_rng(seed)
    permuted_subj_ids = rng.permutation(subj_ids)

    # Split the subject IDs
    split_subj_ids = {}
    curr_split_start = 0
    for k, split in cfg['splits'].items():
        curr_split_end = curr_split_start + split['n']
        split_subj_ids[k] = permuted_subj_ids[curr_split_start:curr_split_end]
        curr_split_start = curr_split_end

    # Parse config for each split
    split_cfgs = parse_dataset_configs(cfg)

    # Create the datasets for each split
    res = {}
    for k, cfg in split_cfgs.items():
        res[k] = sds.dataset.from_slf_dataset(
            slf_ds, cfg, subject_ids=split_subj_ids[k])

    return res


def load_datasets(cfgs: dict[str, Any], seed: int = 42) -> dict[str, tf.data.Dataset]:
    """Load and combine multiple series to tf.data.Dataset."""
    dodh_ds = load_dataset(cfgs['dodh'], seed=seed)
    dodo_ds = load_dataset(cfgs['dodo'], seed=seed)

    datasets = {
        'train': dodh_ds['train'].concatenate(dodo_ds['train']),
        'val': dodh_ds['val'].concatenate(dodo_ds['val']),
        'test': dodh_ds['test'].concatenate(dodo_ds['test']),
    }

    return datasets



"""Convert DOD data from h5 files to sleeplab format."""
import argparse
import h5py
import logging
import numpy as np
import sleeplab_format as slf

from datetime import datetime as dt
from datetime import timedelta
from datetime import timezone
from functools import partial
from pathlib import Path
from zoneinfo import ZoneInfo


logger = logging.getLogger(__name__)


def read_hypnogram(h5: h5py.File, start_ts: dt, epoch_sec: float = 30.0) -> slf.models.Hypnogram:
    """Read the hypnogram from .h5 file and parse to sleeplab format."""
    stage_map = {
        -1: slf.models.AASMSleepStage.UNSCORED,
        0: slf.models.AASMSleepStage.W,
        1: slf.models.AASMSleepStage.N1,
        2: slf.models.AASMSleepStage.N2,
        3: slf.models.AASMSleepStage.N3,
        4: slf.models.AASMSleepStage.R
    }
    h5_hg = h5['hypnogram'][:]
    annotations = []

    for i, h5_stage in enumerate(h5_hg):
        start_delta = i * timedelta(seconds=epoch_sec)
        stage = slf.models.Annotation[slf.models.AASMSleepStage](
            name=stage_map[h5_stage],
            start_ts=start_ts + start_delta,
            start_sec=float(start_delta.seconds),
            duration=epoch_sec
        )
        annotations.append(stage)

    return slf.models.Hypnogram(annotations=annotations, scorer='manual_consensus')


def read_signal_from_h5_file(h5_path: Path, signal_type: str, signal_name: str) -> np.ndarray:
    h5 = h5py.File(h5_path, 'r')
    s = h5['signals'][signal_type][signal_name][:].astype(np.float32)
    return s


def read_sample_arrays(
        h5: h5py.File,
        h5_path: Path,
        start_ts: dt) -> dict[str, slf.models.SampleArray]:
    """Read all signals from a .h5 file and parse to sleeplab format."""
    sample_arrays = {}

    for signal_type in h5['signals']:
        fs = h5['signals'][signal_type].attrs['fs']
        unit = h5['signals'][signal_type].attrs['unit'].decode('utf-8')

        for signal_name in h5['signals'][signal_type]:
            # Use loading functions to avoid reading all data to memory
            s_load_func = partial(read_signal_from_h5_file,
                                  h5_path=h5_path,
                                  signal_type=signal_type,
                                  signal_name=signal_name)
            array_attributes = slf.models.ArrayAttributes(
                name=signal_name,
                start_ts=start_ts,
                sampling_rate=fs,
                unit=unit
            )
            sample_array = slf.models.SampleArray(
                attributes=array_attributes,
                values_func=s_load_func
            )

            sample_arrays[signal_name] = sample_array

    return sample_arrays


def read_subject(h5_path: Path, tz: str) -> slf.models.Subject:
    """Read the signals and hypnogram from .h5 file and parse to sleeplab format Subject.

    Args:
        tz: The timezone as string. The original start_times are as UTC, and
            we want to convert them to correct local times ('US/Pacific' for DOD-O, 'CET' for DOD-H)
    """
    h5 = h5py.File(h5_path, 'r')

    # Presumably, there are no event annotations, so raise an error events are found
    assert len(h5['events']) == 0, f'There are events for {h5_path}'

    # Some recordings are missing start_time, substitute with zero
    try:
        start_ts = (dt
            .fromtimestamp(h5.attrs['start_time'], tz=timezone.utc)
            .astimezone(ZoneInfo(tz))  # Use aware timestamp to specify timezone
            .replace(tzinfo=None)  # Convert to naive timestamp
        )
    except KeyError:
        # If start_time is not found, just use Unix epoch 0
        start_ts = dt.utcfromtimestamp(0)

    sample_arrays = read_sample_arrays(h5, h5_path, start_ts)
    hg = read_hypnogram(h5, start_ts)

    metadata = slf.models.SubjectMetadata(
        subject_id=h5_path.stem,
        recording_start_ts=start_ts
    )

    return slf.models.Subject(
        metadata=metadata,
        sample_arrays=sample_arrays,
        annotations={f'{hg.scorer}_{hg.type}': hg}
    )


def read_to_slf(h5_dir: Path) -> slf.models.Dataset:
    """Parse the whole dataset to sleeplab format Dataset."""
    series = {}

    for subdir in h5_dir.iterdir():
        series_name = subdir.name
        subjects = {}
        h5_paths = [p for p in subdir.iterdir() if p.suffix == '.h5']

        if series_name == 'dodh':
            # DOD-H was recorded in Bretigny-Sur-Orge, France
            tz = 'CET'
        elif series_name == 'dodo':
            # DOD-O was recorded in Stanford, US
            tz = 'US/Pacific'

        for h5_path in h5_paths:
            logger.info(f'Parsing subject data from {h5_path.name}')
            subject = read_subject(h5_path, tz)
            subjects[subject.metadata.subject_id] = subject

        series[series_name] = slf.models.Series(name=series_name, subjects=subjects)

    dataset = slf.models.Dataset(name=h5_dir.name, series=series)

    return dataset


def convert_data(h5_dir: Path, slf_dir: Path, array_format: str = 'numpy', clevel: int = 9) -> None:
    """Run the conversion from .h5 files to sleeplab format."""
    logger.info(f'Reading data to slf Dataset from {h5_dir}')
    dataset = read_to_slf(h5_dir)

    logger.info(f'Writing the dataset to {slf_dir}')
    # The format for annotation files can be 'json' or 'parquet'.
    # array_format can be 'numpy' or 'parquet'. Parquet files will be smaller due to compression.
    slf.writer.write_dataset(dataset, slf_dir, annotation_format='json',
                             array_format=array_format, compression_level=clevel)


def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--src-dir', help='the h5 data dir')
    parser.add_argument('-d', '--dst-dir', help='the directory where the slf dataset is saved')
    parser.add_argument('--array-format', default='numpy', help='Save format of the arrays; `numpy`, `parquet` or `zarr`')
    parser.add_argument('--clevel', type=int, default=9, help='Compression level used with zarr.')

    return parser


if __name__ == '__main__':
    parser = create_parser()
    args = parser.parse_args()
    convert_data(Path(args.src_dir), Path(args.dst_dir), args.array_format, args.clevel)



"""Read files into sleeplab format. The data will be validated while parsing.
"""
import json
import logging
import pandas as pd
import pyarrow.parquet as pq
import zarr

from sleeplab_format.models import *
from pathlib import Path


logger = logging.getLogger(__name__)


JSON_ANNOTATION_SUFFIX = '.a.json'
PARQUET_ANNOTATION_SUFFIX = '.a.parquet'
PARQUET_ANNOTATION_META_SUFFIX = '.a_meta.json'


def read_sample_arrays(subject_dir: Path) -> dict[str, SampleArray] | None:
    """Read all subject's sample arrays.

    Arguments:
        subject_dir: The subject folder.

    Returns:
        All sample arrays in a dictionary.
    """
    sarrs = {}
    for p in subject_dir.iterdir():
        if p.is_dir() and not p.name.startswith('.'):
            with open(p / 'attributes.json', 'rb') as f:
                raw_data = f.read()
                attributes = ArrayAttributes.model_validate_json(raw_data)

            if (p / 'data.npy').exists():
                # Return a function that returns a memmapped numpy array
                values_func = lambda _p=p / 'data.npy': np.load(
                    _p, mmap_mode='r', allow_pickle=False)
            elif (p / 'data.parquet').exists():
                values_func = lambda _p=p / 'data.parquet': pq.read_table(
                    _p)['data'].to_numpy()
            elif (p / 'data.zarr').exists():
                values_func = lambda _p=p / 'data.zarr': zarr.open(_p, mode='r')
            else:
                raise FileNotFoundError(f'No data.npy or data.parquet in {p}')

            assert p.name == attributes.name
            sarrs[p.name] = SampleArray(
                attributes=attributes, values_func=values_func)
    return sarrs


def read_annotations(subject_dir: Path) -> dict[str, list[Annotation]] | None:
    """Read all subject's annotations.

    Arguments:
        subject_dir: The subject folder.

    Returns:
        All annotations in a dictionary.
    """
    annotations = {}
    for p in subject_dir.iterdir():

        if p.name.endswith(JSON_ANNOTATION_SUFFIX):            
            annotation_name = p.name.removesuffix(JSON_ANNOTATION_SUFFIX)
            with open(p, 'rb') as f:
                raw_data = f.read()
                annotations[annotation_name] = BaseAnnotations.model_validate_json(raw_data)
        elif p.name.endswith(PARQUET_ANNOTATION_SUFFIX):
            annotation_name = p.name.removesuffix(PARQUET_ANNOTATION_SUFFIX)
            annotation_meta_path = subject_dir / f'{annotation_name}{PARQUET_ANNOTATION_META_SUFFIX}'

            with open(annotation_meta_path, 'r') as f:
                ann_dict = json.load(f)

            ann_df = pd.read_parquet(p)
            ann_dict['annotations'] = ann_df.to_dict('records')

            annotations[annotation_name] = BaseAnnotations.model_validate(ann_dict)

    if len(annotations) == 0:
        return None
    return annotations


def read_subject(
        subject_dir: Path,
        include_annotations: bool = True) -> Subject:
    """Read a single subject to `sleeplab_format.models.Subject`.

    Arguments:
        subject_dir: The subject folder.
        include_annotations: Whether to include the annotations.

    Returns:
        The resulting subject.
    """
    with open(subject_dir / 'metadata.json', 'rb') as f:
        raw_data = f.read()
        metadata = SubjectMetadata.model_validate_json(raw_data)
    
    sample_arrays = read_sample_arrays(subject_dir)

    if include_annotations:
        annotations = read_annotations(subject_dir)
    else:
        annotations = None

    return Subject(
        metadata=metadata,
        sample_arrays=sample_arrays,
        annotations=annotations,
    )


def read_series(
        series_dir: Path,
        include_annotations: bool = True) -> Series:
    """Read a single series to `sleeplab_format.models.Series`.

    Arguments:
        series_dir: The series root folder.
        include_annotations: Whether to include the annotations.

    Returns:
        The resulting series.
    """
    return Series(
        name=series_dir.name,
        subjects={subject_dir.name: read_subject(
                subject_dir, include_annotations=include_annotations)
            for subject_dir in series_dir.iterdir()
            if not subject_dir.name.startswith('.')}  # Ignore hidden folders.
    )


def read_dataset(
        ds_dir: Path,
        series_names: list[str] | None = None,
        include_annotations: bool = True) -> Dataset:
    """Read a dataset stored in sleeplab-format.

    Arguments:
        ds_dir: The dataset root folder.
        series_names: The series included in the resulting dataset.
        include_annotations: Whether to include annotations or only read the sample arrays.

    Returns:
        The resulting dataset.
    """
    with open(ds_dir / 'metadata.json', 'r') as f:
        ds_meta = json.load(f)

    assert ds_meta['name'] == ds_dir.name
    if ds_meta['version'] != SLEEPLAB_FORMAT_VERSION:
        logger.warning(
            f'Reading dataset version {ds_meta["version"]} with sleeplab-format version {SLEEPLAB_FORMAT_VERSION}')

    if series_names is None:
        series = {series_dir.name: read_series(
                series_dir, include_annotations=include_annotations)
            for series_dir in ds_dir.iterdir()
            if series_dir.is_dir() and not series_dir.name.startswith('.')}  # Ignore hidden folders.
    else:
        series = {series_name: read_series(
                ds_dir / series_name, include_annotations=include_annotations)
            for series_name in series_names}
    
    return Dataset(
        series=series,
        **ds_meta
    )



"""Write validated data into sleeplab format.

The data needs to conform to the types specified in
`sleeplab_format.models`.
"""
import json
import logging
import numcodecs
import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import zarr

from sleeplab_format.models import *
from sleeplab_format.reader import (
    JSON_ANNOTATION_SUFFIX,
    PARQUET_ANNOTATION_SUFFIX,
    PARQUET_ANNOTATION_META_SUFFIX
)
from pathlib import Path


logger = logging.getLogger(__name__)


JSON_INDENT = 2


def write_subject_metadata(
        subject: Subject,
        subject_path: Path) -> None:
    """Write subject metadata to JSON file.
    
    Arguments:
        subject: The sleeplab_format.models.Subject whose metadata will be saved.
        subject_path: The path to the subject folder.
    """
    metadata_path = subject_path / 'metadata.json'
    metadata_path.write_text(
        subject.metadata.model_dump_json(indent=JSON_INDENT, exclude_none=True),
    )


def write_sample_arrays(
        subject: Subject,
        subject_path: Path,
        format: str = 'numpy',
        zarr_chunksize: int | None = 5e6,
        zarr_compression_level: int = 9) -> None:
    """Write all sample arrays of the subject.
    
    Arguments:
        subject: The sleeplab.models.Subject instance.
        subject_path: Path to the folder where the sample arrays are saved.
        format: The save format for the numerical arrays; `numpy`, `parquet` or `zarr`.
        zarr_chunksize: The chunk size in bytes if `format='zarr'`.
        zarr_compression_level: The compression level used with the Zstandard compression.
    """
    for name, sarr in subject.sample_arrays.items():
        assert name == sarr.attributes.name
        sarr_path = subject_path / f'{sarr.attributes.name}'
        sarr_path.mkdir(exist_ok=True)
        
        # Write the attributes
        attr_path = sarr_path / 'attributes.json'
        attr_path.write_text(
            sarr.attributes.model_dump_json(indent=JSON_INDENT, exclude_none=True))

        arr = sarr.values_func()
        if format == 'numpy':
            # Write the array
            arr_fname = 'data.npy'
            np.save(sarr_path / arr_fname, arr, allow_pickle=False)
        elif format == 'zarr':
            arr_fname = 'data.zarr'
            #shuffler = numcodecs.Shuffle(elementsize=4)
            #delta = numcodecs.Delta(dtype='i2')
            #compressor = numcodecs.Blosc(cname='zstd', clevel=5, shuffle=numcodecs.Blosc.NOSHUFFLE)
            #compressor = numcodecs.Blosc(cname='zstd', clevel=9, shuffle=numcodecs.Blosc.NOSHUFFLE)
            #compressor = numcodecs.Blosc(cname='lz4', clevel=5, shuffle=numcodecs.Blosc.NOSHUFFLE)
            #compressor = numcodecs.ZFPY(mode=4, tolerance=1e-3)
            #zarr.save_array(sarr_path / arr_fname, arr, filters=[shuffler, delta], compressor=compressor)
            if zarr_chunksize is not None:
                # Chunk size from bytes to samples
                chunks = (zarr_chunksize // arr.dtype.itemsize,)
                z = zarr.array(arr, chunks=chunks)

            #compressor = numcodecs.Blosc(cname='zstd', clevel=zarr_compression_level, shuffle=numcodecs.Blosc.NOSHUFFLE)
            compressor = numcodecs.Zstd(level=zarr_compression_level)
            zarr.save_array(sarr_path / arr_fname, z, compressor=compressor)
        elif format == 'parquet':
            arr_fname = 'data.parquet'

            # Utilize Arrow to write the data to Parquet file
            arrow_table = pa.Table.from_arrays([arr], names=['data'])

            # Parquet uses Snappy as compression algorithm by default
            pq.write_table(arrow_table, sarr_path / arr_fname)
        else:
            raise AttributeError(f'Unsupported sample array format: {format}')


def write_annotations(
        subject: Subject,
        subject_path: Path,
        format: str = 'json') -> None:
    """Write SLF annotations to disk.
    
    Arguments:
        subject: A sleeplab_format.models.Subject whose annotations will be written.
        subject_path: The path to the subject folder.
        format: The format of annotation files.
    """
    for k, v in subject.annotations.items():
        _msg = f'Annotation key should equal to "{v.scorer}_{v.type}", got "{k}"'
        assert k == f'{v.scorer}_{v.type}', _msg
        
        if format == 'json':
            json_path = subject_path / f'{k}{JSON_ANNOTATION_SUFFIX}'
            json_path.write_text(
                v.model_dump_json(exclude_none=True, indent=JSON_INDENT)
            )
        else:
            # Write the actual annotations in parquet, metadata in json
            metadata_path = subject_path / f'{k}{PARQUET_ANNOTATION_META_SUFFIX}'
            pq_path = subject_path / f'{k}{PARQUET_ANNOTATION_SUFFIX}'
            
            ann_dict = v.model_dump()
            ann_list = ann_dict.pop('annotations')
            
            with open(metadata_path, 'w') as f:
                json.dump(ann_dict, f)
            
            pd.DataFrame(ann_list).to_parquet(pq_path)


def write_subject(
        subject: Subject,
        subject_path: Path,
        annotation_format: str = 'json',
        array_format: str = 'numpy',
        compression_level: int = 9) -> None:
    """Write a single Subject to disk.
    
    Arguments:
        subject: A sleeplab_format.models.Subject to save.
        subject_path: Path to the subject save folder.
        annotation_format: The format of annotation files.
        array_format: The format of the sample array data files.
        compression_level: The zstd compression level if `array_format` is `zarr`.
    """
    subject_path.mkdir(exist_ok=True)
    write_subject_metadata(subject, subject_path)
    
    write_sample_arrays(subject, subject_path, format=array_format,
                        zarr_compression_level=compression_level)

    if subject.annotations is not None:
        write_annotations(subject, subject_path, format=annotation_format)


def write_series(
        series: Series,
        series_path: Path,
        annotation_format: str = 'json',
        array_format: str = 'numpy',
        compression_level: int = 9) -> None:
    """Write a sleeplab_format.models.Series to disk.
    
    Arguments:
        series: The sleeplab_format.models.Series to save.
        series_path: The path to the folder where the series will be saved.
        annotation_format: The format of the annotation files.
        array_format: The format of the sample array data files.
        compression_level: The zstd compression level if `array_format` is `zarr`.
    """
    for sid, subject in series.subjects.items():
        logger.info(f'Writing subject ID {sid}...')
        subject_path = series_path / subject.metadata.subject_id
        write_subject(
            subject,
            subject_path,
            annotation_format=annotation_format,
            array_format=array_format,
            compression_level=compression_level)


def write_dataset(
        dataset: Dataset,
        basedir: str,
        annotation_format: str = 'json',
        array_format: str = 'numpy',
        compression_level: int = 9) -> None:
    """Write a SLF dataset to disk.
    
    Arguments:
        dataset: A sleeplab_format.models.Dataset.
        basedir: The folder where the dataset will be saved.
        annotation_format: The format of the annotation files.
        array_format: The format of the sample array data files.
        compression_level: The zstd compression level if `array_format` is `zarr`.
    """
    assert annotation_format in ['json', 'parquet']
    assert array_format in ['numpy', 'parquet', 'zarr']

    # Create the folder
    dataset_path = Path(basedir) / dataset.name
    logger.info(f'Creating dataset dir {dataset_path}...')
    dataset_path.mkdir(parents=True, exist_ok=True)

    # Write the dataset metadata
    metadata_path = dataset_path / 'metadata.json'
    metadata_path.write_text(
        dataset.model_dump_json(exclude={'series'}, indent=JSON_INDENT))

    # Write the series
    for name, series in dataset.series.items():
        assert name == series.name
        logger.info(f'Writing data for series {series.name}...')
        series_path = dataset_path / series.name
        series_path.mkdir(exist_ok=True)

        write_series(
            series,
            series_path,
            annotation_format=annotation_format,
            array_format=array_format,
            compression_level=compression_level)



import argparse
import dataset
import numpy as np
import tensorflow as tf

from pathlib import Path


def conv_block(x, filters, kernel_size):
    x = tf.keras.layers.Conv1D(
        filters=filters, kernel_size=kernel_size, padding='SAME')(x)
    x = tf.keras.layers.BatchNormalization()(x)
    x = tf.keras.activations.relu(x, alpha=0.1)

    x = tf.keras.layers.Conv1D(
        filters=filters, kernel_size=kernel_size, padding='SAME')(x)
    x = tf.keras.layers.BatchNormalization()(x)
    x = tf.keras.activations.relu(x, alpha=0.1)

    x = tf.keras.layers.MaxPool1D(pool_size=2)(x)

    return x


def create_model(input_names, fs, n_classes=5):
    # Create and concatenate inputs
    inputs = []
    for input_name in input_names:
        inp = tf.keras.layers.Input(shape=(None, 1), name=input_name)
        inputs.append(inp)
    x = tf.keras.layers.Concatenate(name='concat_inputs')(inputs)

    # Create a number of conv blocks to downsample the input resolution from fs to 1Hz
    num_blocks = int(np.log2(fs))
    filters = 8
    kernel_size = 5
    for _ in range(num_blocks):
        x = conv_block(x, filters, kernel_size)
        filters *= 2  # Double the number of features on each block

    # Use avg pooling to downsample the 1Hz signals to 30s epochs
    x = tf.keras.layers.AvgPool1D(pool_size=30)(x)

    # Finally, reduce the channels of the 30-segments to n_classes, and use softmax activation
    x = tf.keras.layers.Conv1D(
        filters=n_classes, kernel_size=1, activation='softmax', padding='SAME', name='hypnogram'
    )(x)

    return tf.keras.Model(inputs=inputs, outputs=[x], name='simple_cnn')


def training_loop(model_dir, epochs, fs=64.0, batch_size=5):
    # Create the datasets
    datasets = dataset.load_datasets(dataset.dataset_config)
    output_names = ['hypnogram']
    input_names = list(set(dataset.component_config.keys()) - set(output_names))

    # Map the datasets to input and output
    def _io_map_func(d):
        inputs = {k: d[k] for k in input_names}
        outputs = {k: tf.expand_dims(d[k], -1) for k in output_names}
        return inputs, outputs

    train_size = len(datasets['train'])
    train_ds = (datasets['train']
        .map(_io_map_func)
        .shuffle(3 * batch_size)
        .batch(batch_size)
        .prefetch(tf.data.AUTOTUNE)
        .repeat()
    )
    val_size = len(datasets['val'])
    val_ds = (datasets['val']
        .map(_io_map_func)
        .batch(1)
        .prefetch(tf.data.AUTOTUNE)
        .repeat()
    )
    test_size = len(datasets['test'])
    test_ds = (datasets['test']
        .map(_io_map_func)
        .batch(1)
        .prefetch(tf.data.AUTOTUNE)
        .repeat()
    )

    # Create the model
    m = create_model(input_names=input_names, fs=fs)
    loss = 'sparse_categorical_crossentropy'
    optimizer = tf.keras.optimizers.experimental.AdamW()
    m.compile(optimizer=optimizer, loss=loss,
              metrics=[tf.keras.metrics.SparseCategoricalAccuracy()])

    # Define training callbacks
    cbs = []
    ckpt_path = Path(model_dir) / 'dod_best_model'
    cbs.append(tf.keras.callbacks.ModelCheckpoint(
        ckpt_path, monitor='val_loss', save_best_only=True
    ))

    # Train the model
    train_steps = train_size // batch_size
    m.fit(
        train_ds,
        steps_per_epoch=train_steps,
        epochs=epochs,
        validation_data=val_ds,
        validation_steps=val_size,
        callbacks=cbs,
        verbose=1
    )

    # Evaluate
    best_m = tf.keras.models.load_model(ckpt_path)
    best_m.evaluate(test_ds, steps=test_size)


def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--model-dir')
    parser.add_argument('--epochs', type=int, default=3)
    return parser


if __name__ == '__main__':
    parser = create_parser()
    args = parser.parse_args()
    training_loop(model_dir=args.model_dir, epochs=args.epochs)



import json
import numpy as np

from pathlib import Path
from sleeplab_format.test_utils.fixtures import *
from sleeplab_format import models


def create_datasets(basedir: Path) -> None:
    ds_name = 'dataset1'
    ds_version = models.SLEEPLAB_FORMAT_VERSION
    series_name = 'series1'
    subject_ids = ['10001', '10002', '10003']
    
    ds_dir = basedir / ds_name
    series_dir = ds_dir / series_name
    series_dir.mkdir(exist_ok=True, parents=True)

    ds_metadata = {'name': ds_name, 'version': ds_version}
    with open(ds_dir / 'metadata.json', 'w') as f:
        json.dump(ds_metadata, f, indent=2)

    for sid in subject_ids:
        subject_dir = series_dir / sid
        subject_dir.mkdir(exist_ok=True)

        subject_metadata_path = subject_dir / 'metadata.json'
        with open(subject_metadata_path, 'w') as f:
            json.dump(subject_metadata(sid), f, indent=2)

        for k, v in sample_arrays().items():
            sarr_path = subject_dir / k
            sarr_path.mkdir(exist_ok=True)

            attr_path = sarr_path / 'attributes.json'
            with open(attr_path, 'w') as f:
                json.dump(v['attributes'], f, indent=2)

            values_path = sarr_path / f'data.npy'
            np.save(values_path, v['values'], allow_pickle=False)

        events_path = subject_dir / 'automatic_aasmevents.a.json'
        with open(events_path, 'w') as f:
            json.dump(events(), f, indent=2)

        hg_path = subject_dir / 'scorer_1_hypnogram.a.json'
        with open(hg_path, 'w') as f:
            json.dump(hypnogram(), f, indent=2)

        study_logs_path = subject_dir / 'study_logs.a.json'
        with open(study_logs_path, 'w') as f:
            json.dump(study_logs(), f, indent=2)


if __name__ == '__main__':
    basedir = Path(__file__).parent / 'datasets'
    create_datasets(basedir)


import numpy as np
import pytest

from sleeplab_format.models import *
from pydantic import ValidationError


def test_samplearray_values(dataset):
    for _, subj in dataset.series['series1'].subjects.items():
        for _, sarr in subj.sample_arrays.items():
            assert np.all(sarr.values_func() == sarr.values)


def test_extra_fields_forbidden(subjects, series):
    with pytest.raises(ValidationError):
        s = Subject(
            metadata=SubjectMetadata(subject_id='123', extra_field='extra'),
            sample_arrays={},
            annotations={}
        )

    with pytest.raises(ValidationError):
        s = Series(name='series', subjects=subjects, extra_field='extra')

    with pytest.raises(ValidationError):
        ds = Dataset(name='ds', series=series, extra_field='extra')



import subprocess

from sleeplab_format import writer
from pathlib import Path


def test_write_dataset(dataset, tmp_path):
    ds_dir = tmp_path / 'datasets'
    writer.write_dataset(dataset, ds_dir)

    # Assert that the created dataset is equal to tests/datasets
    tests_ds_dir = Path(__file__).parent / 'datasets'
    p = subprocess.run(['diff', '-r',
        str(ds_dir.resolve()), str(tests_ds_dir.resolve())])
    assert p.returncode == 0


def test_overwrite_existing_dataset(dataset, tmp_path):
    ds_dir = tmp_path / 'datasets'
    writer.write_dataset(dataset, ds_dir)

    # Assert that the created dataset is equal to tests/datasets
    tests_ds_dir = Path(__file__).parent / 'datasets'
    p = subprocess.run(['diff', '-r',
        str(ds_dir.resolve()), str(tests_ds_dir.resolve())])
    assert p.returncode == 0

    # Write again and assert
    writer.write_dataset(dataset, ds_dir)

    p = subprocess.run(['diff', '-r',
        str(ds_dir.resolve()), str(tests_ds_dir.resolve())])
    assert p.returncode == 0
